@@ -1,7 +1,7 @@
 package Complete::Bash;
 
-our $DATE = '2019-06-26'; # DATE
-our $VERSION = '0.322'; # VERSION
+our $DATE = '2019-06-27'; # DATE
+our $VERSION = '0.323'; # VERSION
 
 use 5.010001;
 use strict;
@@ -397,16 +397,33 @@ sub join_wordbreak_words {
     [$new_words, $cword];
 }
 
-sub _columns {
+sub _terminal_width {
     # XXX need to cache?
     if (eval { require Term::Size; 1 }) {
-        my ($columns, undef) = Term::Size::chars();
-        $columns;
+        my ($cols, undef) = Term::Size::chars();
+        $cols;
     } else {
         $ENV{COLUMNS} // 80;
     }
 }
 
+# given terminal width & number of columns, calculate column width
+sub _column_width {
+    my ($terminal_width, $num_columns) = @_;
+    if (defined $num_columns && $num_columns > 0) {
+        int( ($terminal_width - ($num_columns-1)*2) / $num_columns ) - 1;
+    } else {
+        undef;
+    }
+}
+
+# given terminal width & column width, calculate number of columns
+sub _num_columns {
+    my ($terminal_width, $column_width) = @_;
+    my $n = int( ($terminal_width+2) / ($column_width+2) );
+    $n >= 1 ? $n : 1;
+}
+
 $SPEC{format_completion} = {
     v => 1.1,
     summary => 'Format completion for output (for shell)',
@@ -545,7 +562,7 @@ sub format_completion {
             $msg =~ s/\A +//;
             $msg = " (empty message)" unless length $msg;
         }
-        @res = (sprintf("%-"._columns()."s", $msg), " ");
+        @res = (sprintf("%-"._terminal_width()."s", $msg), " ");
         goto RETURN_RES;
     }
 
@@ -605,26 +622,84 @@ sub format_completion {
         $has_summary = 1 if length $summary;
     }
 
+    my $summary_align = $ENV{COMPLETE_BASH_SUMMARY_ALIGN} // 'left';
+    my $max_columns = $ENV{COMPLETE_BASH_MAX_COLUMNS} // 0;
+    my $terminal_width = _terminal_width();
+    my $column_width = _column_width($terminal_width, $max_columns);
+
+    #warn "terminal_width=$terminal_width, column_width=".($column_width // 'undef')."\n";
+
+  SORT: {
+        # we pre-sort because we want to draw summary lines every N row
+        last unless $summary_align eq 'right';
+        my @orders = sort {
+            # XXX how does bash sort completion entries, exactly? this is still
+            # not right. bash puts -\? between --debug and --format.
+            my $e1 = lc $res[$a]; $e1 =~ s/[^A-Za-z0-9]+//;
+            my $e2 = lc $res[$b]; $e2 =~ s/[^A-Za-z0-9]+//;
+            $e1 cmp $e2;
+        } 0..$#res;
+        @res = map { $res[$_] } @orders;
+        @summaries = map { $summaries[$_] } @orders;
+    }
+
   INSERT_SUMMARIES: {
         last if @res <= 1;
         last unless $has_summary;
         last unless $opts->{show_summaries} //
             $ENV{COMPLETE_BASH_SHOW_SUMMARIES} // 1;
-        my $columns = _columns();
-        my $width = 8;
-        for (@res) {
-            $width = length if $width < length;
+        my $max_entry_width   = 8;
+        my $max_summ_width = 0;
+        for (0..$#res) {
+            $max_entry_width = length $res[$_]
+                if $max_entry_width < length $res[$_];
+            $max_summ_width = length $summaries[$_]
+                if $max_summ_width < length $summaries[$_];
         }
+        #warn "max_entry_width=$max_entry_width, max_summ_width=$max_summ_width\n";
+        if ($summary_align eq 'right') {
+            # if we are aligning summary to the right, we want to fill column
+            # width width
+            if ($max_columns <= 0) {
+                $max_columns = _num_columns(
+                    $terminal_width, ($max_entry_width + 2 + $max_summ_width));
+            }
+            $column_width = _column_width($terminal_width, $max_columns);
+            my $new_max_summ_width = $column_width - 2 - $max_entry_width;
+            $max_summ_width = $new_max_summ_width
+                if $max_summ_width < $new_max_summ_width;
+            #warn "max_columns=$max_columns, column_width=$column_width, max_summ_width=$max_summ_width\n";
+        }
+
+        my $line_every = $ENV{COMPLETE_BASH_SUMMARY_LINE_EVERY} // 4;
         for (0..$#res) {
             my $summary = $summaries[$_];
+            if ($max_columns == 1 &&
+                    $summary_align eq 'right' && ($_+1) % $line_every == 0) {
+                $summary = ("_" x ($max_summ_width - length $summary)) .
+                    $summary;
+            }
             if (length $summary) {
-                $res[$_] = sprintf("%-${width}s -- %s", $res[$_], $summary);
+                $res[$_] = sprintf(
+                    "%-${max_entry_width}s  %".
+                        ($summary_align eq 'right' ? $max_summ_width : '')."s",
+                    $res[$_], $summary);
             }
-            $res[$_] .= " " x ($columns - length($res[$_]))
-                if length($res[$_]) < $columns;
         }
     } # INSERT_SUMMARIES
 
+  MAX_COLUMNS: {
+        last unless $max_columns > 0;
+        my $max_entry_width = 0;
+        for (@res) {
+            $max_entry_width = length if $max_entry_width < length;
+        }
+        last if $max_entry_width >= $column_width;
+        for (@res) {
+            $_ .= " " x ($column_width - length) if $column_width > length;
+        }
+    }
+
   RETURN_RES:
     if ($as eq 'array') {
         return \@res;
@@ -648,7 +723,7 @@ Complete::Bash - Completion routines for bash shell
 
 =head1 VERSION
 
-This document describes version 0.322 of Complete::Bash (from Perl distribution Complete-Bash), released on 2019-06-26.
+This document describes version 0.323 of Complete::Bash (from Perl distribution Complete-Bash), released on 2019-06-27.
 
 =head1 DESCRIPTION
 
@@ -1045,11 +1120,62 @@ Return value:  (any)
 
 =head1 ENVIRONMENT
 
+=head2 COMPLETE_BASH_MAX_COLUMNS
+
+Uint.
+
+Bash will show completion entries in one or several columns, depending on the
+terminal width and the length of the entries (much like a standard non-long
+`ls`). If you prefer completion entries to be shown in a single column no matter
+how wide your terminal is, or how short the entries are, you can set the value
+of this variable to 1. If you prefer a maximum of two columns, set to 2, and so
+on. L</format_completion> will pad the entries with sufficient spaces to limit
+the number of columns.
+
 =head2 COMPLETE_BASH_SHOW_SUMMARIES
 
 Bool. Will set the default for C<show_summaries> option in
 L</format_completion>.
 
+=head2 COMPLETE_BASH_SUMMARY_ALIGN
+
+String. Either C<left> (the default) or C<right>.
+
+The C<left> align looks something like this:
+
+ --bar      Summary about the bar option
+ --baz      Summary about the baz option
+ --foo      Summary about the foo option
+ --schapen  Summary about the schapen option
+
+The C<right> align will make the completion answer look like what you see in the
+B<fish> shell:
+
+ --bar                        Summary about the bar option
+ --baz                        Summary about the baz option
+ --foo                        Summary about the foo option
+ --schapen                Summary about the schapen option
+
+To help match the option and its corresponding summary visually, by default a
+line of underscores is drawn for every 5 lines (configurable via
+L</COMPLETE_BASH_SUMMARY_LINE_EVERY>.
+
+ --bar                        Summary about the bar option
+ --baz                        Summary about the baz option
+ --foo                        Summary about the foo option
+ --foobar   _______________Summary about the foobar option
+ --lam                        Summary about the lam option
+ --qux
+ --quux                                    Alias for --qux
+ --schapen  ______________Summary about the schapen option
+
+=head2 COMPLETE_BASH_SUMMARY_LINE_EVERY
+
+Uint. Default: 4.
+
+Relevant only when L</COMPLETE_BASH_SUMMARY_ALIGN> is set to C<right> (see its
+documentation for more detail).
+
 =head2 COMPLETE_BASH_TRACE
 
 Bool. If set to true, will produce more log statements to L<Log::ger>.
